
def generate_default_config(i: int, column_config: dict = None) -> dict:
  """
  Generate a dictionary with default values for column configurations.

  Args:
      i (int): Index of the column.
      column_config (dict, optional): Dictionary containing column configurations.
        Defaults to None.

  Returns:
      dict: Dictionary with default values for column configurations.
  """

  default_config = {
    'column_name': f"x_{i}",
    'variable_type': 'independent',
    'probability_distribution': {
      'type': 'uniform',
      'parameter': {
        'low': 0,
        'high': 1
      }
    }
  }

  # if column_info is not None, then update default_config with column_info
  if column_config is not None:
    default_config.update(column_config)

  return default_config

class DnxmyConfig:
  def __init__(self):
    """
    Initialize the DnxmyConfig class.

    Returns:
        DnxmyConfig: DnxmyConfig class.
    """

    # initialize the class attributes
    self.dataset_config: list = []
    self.missing_config: list = []


  def add_independent_column(self, col_name: str, probability_distribution_type: str = 'uniform', probability_distribution_params: dict = {'low': 0, 'high': 1}):
    """
    Add a column configuration for an independent variable.

    Args:
        col_name (str): Name of the column.
        probability_distribution_type (str): Type of the probability distribution.
          Defaults to 'uniform'.
        probability_distribution_params (dict, optional): Parameters of the probability distribution.
          Defaults to {'low': 0, 'high': 1}.
    """
    # create the configuration for the column
    self.dataset_config.append({
      'column_name': col_name,
      'variable_type': 'independent',
      'probability_distribution': {
        'type': probability_distribution_type,
        'parameter': probability_distribution_params
      }
    })


  def add_categorical_column(self, col_name: str, value: list, probability: list):
    """
    Add a column configuration for a categorical variable.

    Args:
        col_name (str): Name of the column.
        value (list): List of values of the categorical variable.
        probability (list): List of probabilities of the values.
    """
    categories = []
    for i in range(len(value)):
      categories.append({
        'value': value[i],
        'probability': probability[i]
      })

    # add a column configuration for the categorical variable
    self.dataset_config.append({
      'column_name': col_name,
      'variable_type': 'independent',
      'probability_distribution': {
        'type': 'categorical',
        'parameter': {
          'categories': categories
        }
      }
    })


  def add_constant_column(self, col_name: str, constant_value: float = 1):
    """
    Add a column configuration for a constant variable.

    Args:
        col_name (str): Name of the column.
        constant_value (float): Value of the constant.
          Defaults to 1.
    """
    # add a column configuration for the constant variable
    self.dataset_config.append({
      'column_name': col_name,
      'variable_type': 'constant',
      'constant_value': constant_value
    })


  def add_time_part_column(self, col_name: str, start_time: str, time_unit: str, time_format: str = '%Y-%m-%d'):
    """
    Add a column configuration for a time part variable.

    Args:
        col_name (str): Name of the column.
        start_time (str): Start time of the time variable.
        time_unit (str): Time unit of the time variable.
        time_format (str): Time format of the time variable.
          Defaults to '%Y-%m-%d'.
    """
    # add a column configuration for the time variable
    self.dataset_config.append({
      'column_name': col_name,
      'variable_type': 'time_part',
      'time_part_config': {
        'start_time': start_time,
        'time_unit': time_unit,
        'time_format': time_format
      }
    })


  def add_arma_column(self, 
                      col_name:str, 
                      intercept: float = 0, 
                      sigma: float = 0, 
                      ar_initial: float = None, 
                      ar_order: int = None, 
                      ar_params: list = None, 
                      ar_shock_time: list = None,
                      ar_shock_type: list = None, 
                      ar_shock_value: list = None, 
                      ma_initial: float = None, 
                      ma_order: int = None,
                      ma_params: list = None,
                      ma_shock_time: list = None,
                      ma_shock_type: list = None, 
                      ma_shock_value: list = None):
    """
    Add a column configuration for an ARMA variable.

    Args:
        col_name (str): Name of the column.
        intercept (float): Intercept for the ARMA variable.
          Defaults to 0.
        sigma (float): Sigma for the ARMA variable.
          Defaults to 0.
        ar_initial (float): Initial value for the AR part of the ARMA variable.
          Defaults to None.
        ar_order (int): Order of the AR part of the ARMA variable.
          Defaults to None.
        ar_params (list): List of AR parameters for the ARMA variable.
          Defaults to None.
        ar_shock_type (list): List of shock types for the AR part of the ARMA variable.
          Defaults to None.
        ar_shock_value (list): List of shock values for the AR part of the ARMA variable.
          Defaults to None.
        ma_initial (float): Initial value for the MA part of the ARMA variable.
          Defaults to None.
        ma_order (int): Order of the MA part of the ARMA variable.
          Defaults to None.
        ma_params (list): List of MA parameters for the ARMA variable.
          Defaults to None.
        ma_shock_type (list): List of shock types for the MA part of the ARMA variable.
          Defaults to None.
        ma_shock_value (list): List of shock values for the MA part of the ARMA variable.
          Defaults to None.
    """
    # create a dictionary for the AR shocks
    ar_shock_dict = {}
    for i, t, v in zip(ar_shock_time, ar_shock_type, ar_shock_value):
      ar_shock_dict[i] = {}
      ar_shock_dict[i]['type'] = t
      ar_shock_dict[i]['value'] = v
    
    # create a dictionary for the MA shocks
    ma_shock_dict = {}
    for i, t, v in zip(ma_shock_time, ma_shock_type, ma_shock_value):
      ma_shock_dict[i] = {}
      ma_shock_dict[i]['type'] = t
      ma_shock_dict[i]['value'] = v

    # create the column configuration dictionary
    self.dataset_config.append({
      'column_name': col_name,
      'variable_type': 'time_series',
      'time_series_config': {
        'intercept': intercept,
        'sigma': sigma,
        'ar_params': {
          'initial': ar_initial,
          'order': ar_order,
          'params': ar_params,
          'shock': ar_shock_dict
        },
        'ma_params': {
          'initial': ma_initial,
          'order': ma_order,
          'params': ma_params,
          'shock': ma_shock_dict
        }
      }
    })


  def add_dependent_colmn(self, 
                          col_name: str, 
                          variables: list, 
                          beta: list, 
                          intercept: float, 
                          offset_column: str = None, 
                          offset_function: str = 'default', 
                          link_function: str = 'identity'):
    """
    Add a column configuration for a dependent variable.

    Args:
        col_name (str): Name of the column.
        variables (list): List of columns on which the variable depends.
        beta (list): List of coefficients for the dependent variable.
        intercept (float): Intercept for the dependent variable.
        offset_column (str): Name of the column to use for the offset.
          Defaults to None.
        offset_function (str): Function to use for the offset.
          Defaults to 'default'.
        link_function (str): Link function to use for the dependent variable.
          Defaults to 'identity'.
    """
    # add a column configuration for the dependent variable
    self.dataset_config.append({
      'column_name': col_name,
      'variable_type': 'dependent',
      'dependent_on': {
        'variables': variables,
        'beta': beta,
        'intercept': intercept,
        'offset': {
          'column_name': offset_column,
          'function': offset_function
        },
        'link_function': link_function
      }
    })


  def t_sort(self):
    """
    Optimize the order of dataset configuration using topological sorting.
    """
    sorted_list = []
    dependency_dict = {}
    column_dict = {}

    # create a dictionary of dependencies
    for col in self.dataset_config:
      col_name = col['column_name']
      if col['variable_type'] == 'dependent':
        dependency_dict[col_name] = col['dependent_on']['variables'].copy()
        if col['dependent_on'].get('offset')['column_name'] is not None:
          dependency_dict[col_name].append(col['dependent_on']['offset']['column_name'])
      else:
        dependency_dict[col_name] = []

    # sort the columns based on dependencies
    while len(dependency_dict) != 0:
      independent_cols = [col for col, dep_cols in dependency_dict.items() if len(dep_cols) == 0]

      # if there is no independent columns, there is a circular dependency!
      if len(independent_cols) == 0:
        raise ValueError("Circular dependency detected!")

      sorted_list.extend(independent_cols)
      for col in independent_cols:
        del dependency_dict[col]
        for dep_cols in dependency_dict.values():
          for dep_col in dep_cols:
            if col == dep_col:
              dep_cols.remove(col)

    column_dict = {col['column_name']: col for col in self.dataset_config}
    self.dataset_config = [column_dict[col] for col in sorted_list if col in column_dict]


  def set_dataset_config(self, m: int):
    """
    Set dataset configurations based on the provided configurations or the default configurations.
    """
    for i in range(m):
      if i < len(self.dataset_config):
        column_config = self.dataset_config[i]
      else:
        column_config = None
        self.dataset_config.append(column_config)

      self.dataset_config[i] = generate_default_config(i, column_config)


  def add_missing_config(self, target_col_name: str, missing_type: str = 'MCAR', missing_rate: float = None, dependent_on: str = None):
    """
    Generate a missing configuration dictionary.

    Args:
        missing_type (str): Type of missingness.
          Defaults to 'MCAR'.
        target_column_name (str): Name of the column to which the missingness is applied.
        missing_rate (float): Missing rate.
          Defaults to None.
        dependent_on (str): String expressing a missing condition in query string format in pandas.DataFrame.query.
          Defaults to None.

    Returns:
        dict: Dictionary of missing configurations.
    """
    if missing_type not in ['MCAR', 'MAR', 'MNAR']:
      raise ValueError("Missing type must be one of 'MCAR', 'MAR', or 'MNAR'!")
    
    # add a missing configuration
    self.missing_config.append({
      'missing_type': missing_type,
      'target_column_name': target_col_name,
      'missing_params': {
        'missing_rate': missing_rate,
        'dependent_on': dependent_on
      }
    })


# import libraries
import numpy as np
import pandas as pd

# list of supported probability distributions
DISTRIBUTION_LIST = ['uniform', 'normal', 'beta', 'gamma', 'exponential', 'poisson', 'binomial', 'lognormal', 'chisquare', 'f', 't', 'multivariate_normal', 'multinomial', 'dirichlet', 'laplace', 'logistic', 'logseries', 'negative_binomial', 'noncentral_chisquare', 'noncentral_f', 'pareto', 'rayleigh', 'standard_cauchy', 'standard_exponential', 'standard_gamma', 'standard_normal', 'standard_t', 'triangular', 'vonmises', 'wald', 'weibull', 'zipf', 'gunbel', 'hypergeometric']


def generate_random_samples(col_name: str, probability_distribution: dict, n: int) -> pd.Series:
  """
  Generate random samples based on the provided probability distribution.

  Args:
      col_name (str): Name of the column.
      probability_distribution (dict): Dictionary containing the probability distribution.
      n (int): Number of samples to be generated.

  Returns:
      pd.Series: Series containing the generated samples.
  """

  distribution_type = probability_distribution.get('type')
  params = probability_distribution.get('parameter')
  if distribution_type is None or params is None:
    raise ValueError("distribution_type and params must be specified.")

  if distribution_type == 'categorical':
    categories, probabilities = zip(*[(c['value'], c['probability']) for c in params['categories']])
    probabilities = np.array(probabilities) / sum(probabilities)
    return pd.Series(np.random.choice(categories, n, p=probabilities), name=col_name)

  if distribution_type not in DISTRIBUTION_LIST:
    raise ValueError(f"distribution_type must be in {DISTRIBUTION_LIST}")

  kwargs = {k: params[k] for k in params if k != 'loc'}
  if 'loc' in params:
    kwargs['loc'] = params['loc']

  if distribution_type == 'multivariate_normal':
    kwargs['cov'] = np.array(kwargs['cov'])

  if distribution_type == 'categorical':
    kwargs['p'] = probabilities

  # generate random samples
  return pd.Series(getattr(np.random, distribution_type)(**kwargs, size=n), name=col_name)


def generate_time_part(col_name: str, time_part_config: dict, n: int) -> pd.Series:
  """
  Generate a time series based on the provided time configurations.

  Args:
      col_name (str): Name of the column.
      time_part_config (dict): Dictionary containing the time configurations.
      n (int): Number of samples to generate.

  Returns:
      pd.Series: Series containing the generated time series.
  """
  if not time_part_config:
    raise ValueError("time_part_config must be specified.")
  
  start_time, time_unit, time_format = time_part_config.get('start_time'), time_part_config.get('time_unit'), time_part_config.get('time_format')
  if not all([start_time, time_unit, time_format]):
    raise ValueError("start_time, time_unit, and time_format must be specified in time_part_config.")
  
  # generate time series
  time_series = pd.Series(pd.date_range(start_time, periods=n, freq=time_unit).strftime(time_format), name = col_name)

  return time_series


def generate_arma_samples(col_name: str, time_series_config: dict, n: int) -> pd.Series:
  """
  Generate a time series(ARMA model) based on the provided time configurations.

  Args:
      col_name (str): Name of the column.
      time_series_config (dict): Dictionary containing the time configurations.
      n (int): Number of samples to generate.

  Returns:
      pd.Series: Series containing the generated samples.
  """
  if time_series_config is None:
    raise ValueError("time_series_config must be specified.")
  
  ar_params = time_series_config.get('ar_params', {})
  ma_params = time_series_config.get('ma_params', {})
  
  ar_samples = [ar_params.get('initial', 0)]
  ma_samples = [ma_params.get('initial', 0)]
  
  # generate initial samples
  arma_samples = [ar_samples[0] + ma_samples[0] + time_series_config['intercept'] + np.random.normal(0, time_series_config['sigma'])]
  
  # generate ARMA samples
  for i in range(1, n):
    ar = sum([ar_params['params'][j] * ar_samples[i - j - 1] for j in range(min(i, ar_params.get('order', 0)))])
    
    # add shock
    if ar_params['shock'].get(i) is not None:
      shock_value = ar_params['shock'][i]['value']
      if ar_params['shock'][i]['type'] == 'sigma':
        shock_value *= time_series_config['sigma']
      ar += shock_value
    ar_samples.append(ar)

    ma = sum([ma_params['params'][j] * ma_samples[i - j - 1] for j in range(min(i, ma_params.get('order', 0)))])

    # add shock
    if ma_params['shock'].get(i) is not None:
      shock_value = ma_params['shock'][i]['value']
      if ma_params['shock'][i]['type'] == 'sigma':
        shock_value *= time_series_config['sigma']
      ma += shock_value
    ma_samples.append(ma)

    arma_samples.append(ar + ma + np.random.normal(0, time_series_config['sigma']))
  
  return pd.Series(arma_samples, name=col_name)


def generate_dependent_samples(col_name: str, dataset_config: list, df, dependent_on: dict, n: int, offset: dict = None) -> pd.Series:
  """
  Generate dependent samples based on the provided dependent_on configurations.

  Args:
      col_name (str): Name of the column.
      dataset_config (list): List containing the dataset configurations.
      df (pd.DataFrame): Dataframe containing the dataset.
      dependent_on (dict): Dictionary containing the dependent_on configurations.
      n (int): Number of samples to generate.
      offset (dict, optional): Dictionary containing the offset configurations. 
        Defaults to None.

  Returns:
      pd.Series: Series containing the generated samples.
  """

  intercept = dependent_on['intercept']
  link_function = dependent_on['link_function']
  beta = dependent_on['beta']
  sigma = dependent_on.get('sigma') or 0
  offset = dependent_on.get('offset')
  offset_function = offset['function'] if offset is not None else 'default'

  column_names = dependent_on['variables']

  # transform data  
  variables = df[column_names].to_numpy()
  beta = np.array(beta)
  intercept = np.full(n, intercept)
  offset = df[offset['column_name']].to_numpy() if offset['column_name'] is not None else np.zeros(n)

  offset_functions = {
    'default': lambda x: x,
    'log': lambda x: np.log(x)
  }

  # define link function
  link_functions = {
    'identity': lambda x: x,
    'logit': lambda x: 1 / (1 + np.exp(-x)),
    'probit': lambda x: 0.5 * (1 + erf(x / np.sqrt(2))),
    'cloglog': lambda x: 1 - np.exp(-np.exp(x)),
    'log': lambda x: np.exp(x),
    'sqrt': lambda x: np.sqrt(x),
    'inverse': lambda x: 1 / x,
    'loglog': lambda x: np.exp(-np.exp(-x)),
    'cauchy': lambda x: 1 / (1 + x ** 2),
    'logc': lambda x: np.log(x),
    'power': lambda x: x ** 2,
    'logloglog': lambda x: np.exp(-np.exp(-np.exp(x))),
    'loglogloglog': lambda x: np.exp(-np.exp(-np.exp(-np.exp(x)))),
    'logloglogloglog': lambda x: np.exp(-np.exp(-np.exp(-np.exp(-np.exp(x)))))
  }
  # generate samples
  return pd.Series(link_functions[link_function](np.dot(variables, beta) + offset_functions[offset_function](offset) + intercept) + np.random.normal(0, sigma, n), name = col_name)


import numpy as np
import pandas as pd

from .config_generator import DnxmyConfig
from .variable_generator import generate_random_samples, generate_time_part, generate_arma_samples, generate_dependent_samples


class Dnxmy:
  def __init__(self, n: int, m: int = None, dnxmy_config: DnxmyConfig = None, seed: int = 0):
    """
    A class for generating random datasets based on column configurations.

    Args:
        n (int): Number of samples in the generated dataset.
        m (int, optional): Number of columns in the generated dataset.
          Defaults to None.
        dnxmy_config (DnxmyConfig, optional): Instance of the DnxmyConfig class.
          Defaults to None.
        seed (int, optional): Seed for the random number generator. Defaults to 0.

    Returns:
        Dnxmy: An instance of the Dnxmy class.
    """

    # initialize the class attributes
    self.n: int = n
    self.seed: int = seed
    self.m: int = m
    self.dnxmy_config: dnxmy_config = dnxmy_config or DnxmyConfig()


  def generate(self) -> pd.DataFrame:
    """
    Generate the data based on the provided configurations.
    
    Returns:
        pd.DataFrame: DataFrame containing the generated data.
    """
    # set the seed
    if self.seed is not None:
      np.random.seed(self.seed)

    if self.m < len(self.dnxmy_config.dataset_config):
      self.m = len(self.dnxmy_config.dataset_config)

    # set the column configurations
    self.dnxmy_config.set_dataset_config(self.m)
    self.dnxmy_config.t_sort()
    self.dataset_config = self.dnxmy_config.dataset_config

    column_names = [column_config['column_name'] for column_config in self.dataset_config]
    
    # generate the data
    self.df = pd.DataFrame()

    for column_config in self.dataset_config:
      if column_config['variable_type'] == 'independent':
        self.df = pd.concat([self.df, generate_random_samples(column_config['column_name'], column_config['probability_distribution'], self.n)], axis=1)
      elif column_config['variable_type'] == 'constant':
        self.df = pd.concat([self.df, pd.Series([column_config['constant_value']] * self.n, name=column_config['column_name'])], axis=1)
      elif column_config['variable_type'] == 'time_part':
        self.df = pd.concat([self.df, generate_time_part(column_config['column_name'], column_config['time_part_config'], self.n)], axis=1)
      elif column_config['variable_type'] == 'time_series':
        self.df = pd.concat([self.df, generate_arma_samples(column_config['column_name'], column_config['time_series_config'], self.n)], axis=1)
      elif column_config['variable_type'] == 'dependent':
        self.df = pd.concat([self.df, generate_dependent_samples(column_config['column_name'], self.dataset_config, self.df, column_config['dependent_on'], self.n)], axis=1)

    return self.df


  def add_samples(self, n: int) -> pd.DataFrame:
    """
    Add samples to the generated data.

    Args:
        n (int): Number of samples to be added.

    Returns:
        pd.DataFrame: DataFrame containing the generated data with added samples.
    """
    if self.df is None:
      raise Exception('Data has not been generated yet')

    self.df_add = pd.DataFrame()

    for column_config in self.dataset_config:
      if column_config['variable_type'] == 'independent':
        self.df_add = pd.concat([self.df_add, generate_random_samples(column_config['column_name'], column_config['probability_distribution'], n)], axis=1)
      elif column_config['variable_type'] == 'constant':
        self.df_add = pd.concat([self.df_add, pd.Series([column_config['constant_value']] * n, name=column_config['column_name'])], axis=1)
      elif column_config['variable_type'] == 'time_part':
        self.df_add = pd.concat([self.df_add, generate_time_part(column_config['column_name'], column_config['time_part_config'], n)], axis=1)
      elif column_config['variable_type'] == 'time_series':
        self.df_add = pd.concat([self.df_add, generate_arma_samples(column_config['column_name'], column_config['time_series_config'], n)], axis=1)
      elif column_config['variable_type'] == 'dependent':
        self.df_add = pd.concat([self.df_add, generate_dependent_samples(column_config['column_name'], self.dataset_config, self.df_add, column_config['dependent_on'], n)], axis=1)

    # add the new samples to the existing data
    self.df = pd.concat([self.df, self.df_add], ignore_index=True)
    
    # update the number of samples
    self.n += n

    return self.df


  def miss(self) -> pd.DataFrame:
    """
    Generate missing values based on the provided configurations.
    
    Returns:
        pd.DataFrame: DataFrame containing the generated data with missing values.
    """
    df_missing = self.df.copy()

    # set the seed
    if self.seed is not None:
      np.random.seed(self.seed)
    
    # set the missing configurations
    self.missing_config = self.dnxmy_config.missing_config
    
    for missing_column_config in self.missing_config:
      missing_type = missing_column_config.get('missing_type', None)
      target_column_name = missing_column_config.get('target_column_name', None)
      missing_rate = missing_column_config.get('missing_params', {}).get('missing_rate', None)
      dependent_on = missing_column_config.get('missing_params', {}).get('dependent_on', None)

      # missing
      if missing_type == 'MCAR':
        missing_index = np.random.choice(self.n, int(self.n * missing_rate), replace=False)
        df_missing.loc[missing_index, target_column_name] = np.nan
      elif missing_type == 'MAR':
        missing_index = df_missing.query(dependent_on).index
        missing_index = np.random.choice(missing_index, int(len(missing_index) * missing_rate), replace=False)
        df_missing.loc[missing_index, target_column_name] = np.nan
      elif missing_type == 'MNAR':
        df_missing.loc[df_missing.query(dependent_on).index, target_column_name] = np.nan
      else:
        raise Exception('Invalid missing type')
    print(df_missing)
    return df_missing, self.df





# Author: Yuki Yamamoto <curious.yamamon@gmail.com>
# Copyright (c) 2023 Yuki Yamamoto
# License: MIT License

from setuptools import setup

import dnxmy

DESCRIPTION = "dnxmy: dummy data generator for machine learning and statistics"
NAME = 'dnxmy'
AUTHOR = 'Yuki Yamamoto'
AUTHOR_EMAIL = 'curious.yamamon@gmail.com'
URL = 'https://github.com/roseiricho/dnxmy'
LICENSE = 'MIT License'
DOWNLOAD_URL = 'https://github.com/roseiricho/dnxmy'
VERSION = dnxmy.__version__
PYTHON_REQUIRES = ">=3.6"

INSTALL_REQUIRES = [
    'numpy>=1.22.4',
    'pandas>=1.4.4'
]

PACKAGES = [
    'dnxmy'
]

CLASSIFIERS = [
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3 :: Only',
    'Topic :: Scientific/Engineering'
]

with open('README.md', encoding='UTF-8') as fp:
    readme = fp.read()
with open('CONTACT.txt', 'r') as fp:
    contacts = fp.read()
long_description = readme + '\n\n' + contacts
long_description_content_type = 'text/markdown'

setup(name=NAME,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      maintainer=AUTHOR,
      maintainer_email=AUTHOR_EMAIL,
      description=DESCRIPTION,
      long_description=long_description,
      long_description_content_type=long_description_content_type,
      license=LICENSE,
      url=URL,
      version=VERSION,
      download_url=DOWNLOAD_URL,
      python_requires=PYTHON_REQUIRES,
      install_requires=INSTALL_REQUIRES,
      packages=PACKAGES,
      classifiers=CLASSIFIERS
    )